from __future__ import annotations

from datetime import datetime
from typing import Any, List, Optional

from pydantic import BaseModel, validator

from ikea_api._constants import Constants
from ikea_api.wrappers import types
from ikea_api.wrappers._parsers import translate_from_dict
from ikea_api.wrappers._parsers.item_base import ItemCode

__all__ = ["main"]

DELIVERY_TYPES = {
    "ru": {
        "HOME_DELIVERY": "Доставка",
        "PUP": "Пункт самовывоза",
        "PUOP": "Магазин",
        "CLICK_COLLECT_STORE": "Магазин",
        "MOCKED_CLICK_COLLECT_STORE": "Магазин",
        "IBES_CLICK_COLLECT_STORE": "Магазин",
    }
}

SERVICE_TYPES = {
    "ru": {
        "CURBSIDE": "без подъёма",
        "STANDARD": None,
        "EXPRESS": "экспресс",
        "EXPRESS_CURBSIDE": "без подъёма экспресс",
    }
}

SERVICE_PROVIDERS = {
    "ru": {
        "DPD": "DPD",
        "BUSINESSLINES": "Деловые линии",
        "russianpost": "Почта России",
        "OZON": "Ozon",
    }
}

#
# Shared
#


class SolutionPrice(BaseModel):
    inclTax: int


class EarliestPossibleSlot(BaseModel):
    fromDateTime: datetime


class TimeWindows(BaseModel):
    # TODO: Check if this is the same for home delivery services
    earliestPossibleSlot: Optional[EarliestPossibleSlot]


class SelectableInfo(BaseModel):
    selectable: bool

    @validator("selectable", pre=True)
    def validate_selectable(cls, v: Any):
        return v == "YES"


class Metadata(BaseModel):
    selectableInfo: SelectableInfo


class UnavailableItem(BaseModel):
    itemNo: ItemCode
    availableQuantity: int


def get_date(deliveries: list[HomeDelivery] | None):
    if not deliveries:
        return

    for delivery in deliveries:
        if delivery.timeWindows and delivery.timeWindows.earliestPossibleSlot:
            return delivery.timeWindows.earliestPossibleSlot.fromDateTime


def get_type(service: HomeDeliveryService | CollectDeliveryService):
    delivery_type = translate_from_dict(DELIVERY_TYPES, service.fulfillmentMethodType)
    if service.solution:
        service_type = translate_from_dict(SERVICE_TYPES, service.solution)
        if service_type:
            return f"{delivery_type} {service_type}"
    return delivery_type


def get_price(service: HomeDeliveryService | CollectDeliveryService):
    if service.solutionPrice:
        return service.solutionPrice.inclTax
    return 0


def get_unavailable_items(
    service: HomeDeliveryService | CollectDeliveryService,
) -> list[types.UnavailableItem]:
    if not service.unavailableItems:
        return []
    return [
        types.UnavailableItem(item_code=i.itemNo, available_qty=i.availableQuantity)
        for i in service.unavailableItems
    ]


#
# Home Delivery Services
#


class HomeDelivery(BaseModel):
    type: str
    timeWindows: Optional[TimeWindows]


class HomePossibleDeliveries(BaseModel):
    deliveries: List[HomeDelivery]


class HomeDeliveryService(BaseModel):
    metadata: Metadata
    fulfillmentMethodType: str
    solution: Optional[str]
    solutionPrice: Optional[SolutionPrice]
    possibleDeliveries: Optional[HomePossibleDeliveries]
    unavailableItems: Optional[List[UnavailableItem]]


class HomePossibleDeliveryServices(BaseModel):
    deliveryServices: List[HomeDeliveryService]


class HomeDeliveryServicesResponse(BaseModel):
    possibleDeliveryServices: Optional[HomePossibleDeliveryServices]


def parse_home_delivery_services(response: dict[str, Any]):
    parsed_response = HomeDeliveryServicesResponse(**response)

    res: list[types.DeliveryService] = []
    if not parsed_response.possibleDeliveryServices:
        return res

    for service in parsed_response.possibleDeliveryServices.deliveryServices:
        if not (service.possibleDeliveries and service.possibleDeliveries.deliveries):
            continue

        is_available = service.metadata.selectableInfo.selectable
        unavailable_items = get_unavailable_items(service)

        if not is_available and not unavailable_items:
            continue

        res.append(
            types.DeliveryService(
                is_available=is_available,
                date=get_date(service.possibleDeliveries.deliveries),
                type=get_type(service),
                price=get_price(service),
                service_provider=None,
                unavailable_items=unavailable_items,
            )
        )
    return res


#
# Collect Delivery Services
#


class PickUpPoint(BaseModel):
    metadata: Metadata
    timeWindows: Optional[TimeWindows]
    identifier: Optional[str]


class PossiblePickUpPoints(BaseModel):
    pickUpPoints: List[PickUpPoint]


class CollectDelivery(BaseModel):
    type: str
    possiblePickUpPoints: PossiblePickUpPoints


class CollectPossibleDeliveries(BaseModel):
    deliveries: List[CollectDelivery]


class CollectDeliveryService(BaseModel):
    fulfillmentMethodType: str
    solution: Optional[str]
    solutionPrice: Optional[SolutionPrice]
    possibleDeliveries: Optional[CollectPossibleDeliveries]
    unavailableItems: Optional[List[UnavailableItem]]


class CollectPossibleDeliveryServices(BaseModel):
    deliveryServices: List[CollectDeliveryService]


class CollectDeliveryServicesResponse(BaseModel):
    possibleDeliveryServices: Optional[CollectPossibleDeliveryServices]


def get_service_provider(pickup_point: PickUpPoint):
    identifier = pickup_point.identifier
    if not identifier:
        return
    for provider, pretty_name in SERVICE_PROVIDERS.get(
        Constants.LANGUAGE_CODE, {}
    ).items():
        if provider in identifier:
            return pretty_name
    return identifier


def parse_collect_delivery_services(response: dict[str, Any]):
    parsed_response = CollectDeliveryServicesResponse(**response)

    res: list[types.DeliveryService] = []
    if not parsed_response.possibleDeliveryServices:
        return res

    for service in parsed_response.possibleDeliveryServices.deliveryServices:
        if not service.possibleDeliveries:
            continue

        type_ = get_type(service)
        price = get_price(service)
        unavailable_items = get_unavailable_items(service)

        for delivery in service.possibleDeliveries.deliveries:
            for point in delivery.possiblePickUpPoints.pickUpPoints:
                date = (
                    point.timeWindows.earliestPossibleSlot.fromDateTime
                    if point.timeWindows and point.timeWindows.earliestPossibleSlot
                    else None
                )

                is_available = point.metadata.selectableInfo.selectable

                if not is_available and not unavailable_items:
                    continue

                res.append(
                    types.DeliveryService(
                        is_available=is_available,
                        date=date,
                        type=type_,
                        price=price,
                        service_provider=get_service_provider(point),
                        unavailable_items=unavailable_items,
                    )
                )

    return res


def main(
    *,
    home_delivery_services_response: dict[str, Any],
    collect_delivery_services_response: dict[str, Any],
) -> list[types.DeliveryService]:
    return parse_home_delivery_services(
        home_delivery_services_response
    ) + parse_collect_delivery_services(collect_delivery_services_response)



from __future__ import annotations

import datetime
from typing import Any, Callable, List, Optional, TypeVar

from pydantic import BaseModel as PydanticBaseModel
from pydantic import HttpUrl
from pydantic.fields import Field, FieldInfo
from pydantic.main import ModelMetaclass as PydanticModelMetaclass

_T = TypeVar("_T")


# https://github.com/microsoft/pyright/blob/main/specs/dataclass_transforms.md#applying-to-pydantic
def __dataclass_transform__(
    *,
    eq_default: bool = True,
    order_default: bool = False,
    kw_only_default: bool = False,
    field_descriptors: tuple[type | Callable[..., Any], ...] = (()),
) -> Callable[[_T], _T]:
    return lambda a: a


@__dataclass_transform__(kw_only_default=True, field_descriptors=(Field, FieldInfo))
class ModelMetaclass(PydanticModelMetaclass):
    pass


class BaseModel(PydanticBaseModel, metaclass=ModelMetaclass):
    pass


class ChildItem(BaseModel):
    name: Optional[str]
    item_code: str
    weight: float
    qty: int


class ParsedItem(BaseModel):
    is_combination: bool
    item_code: str
    name: str
    image_url: Optional[str]
    weight: float
    child_items: List[ChildItem]
    price: int
    url: str
    category_name: Optional[str]
    category_url: Optional[HttpUrl]


class IngkaItem(BaseModel):
    is_combination: bool
    item_code: str
    name: str
    image_url: Optional[str]
    weight: float
    child_items: List[ChildItem]


class PipItem(BaseModel):
    item_code: str
    price: int
    url: str
    category_name: Optional[str]
    category_url: Optional[HttpUrl]


class UnavailableItem(BaseModel):
    item_code: str
    available_qty: int


class DeliveryService(BaseModel):
    is_available: bool
    date: Optional[datetime.date]
    type: str
    price: int
    service_provider: Optional[str]
    unavailable_items: List[UnavailableItem]


class GetDeliveryServicesResponse(BaseModel):
    delivery_options: List[DeliveryService]
    cannot_add: List[str]


class CostsOrder(BaseModel):
    delivery_cost: float
    total_cost: float


class StatusBannerOrder(BaseModel):
    purchase_date: datetime.date
    delivery_date: datetime.date


class PurchaseInfo(StatusBannerOrder, CostsOrder):
    pass


class PurchaseHistoryItem(BaseModel):
    id: str
    status: str
    price: int
    datetime: str
    datetime_formatted: str
    store: str


CannotAddItems = List[str]



from __future__ import annotations

import re

import requests


def _get_unshortened_links_from_ingka_pagelinks(message: str):
    session = requests.Session()
    base_url = "https://ingka.page.link/"

    for postfix in re.findall("ingka.page.link/([0-9A-z]+)", message):
        resp = session.get(base_url + postfix, allow_redirects=False)
        location_header = resp.headers.get("Location")
        if location_header is not None:
            yield location_header


def parse_item_codes(
    item_codes: str | list[str], unshorten_ingka_pagelinks: bool = False
) -> list[str]:
    if item_codes == []:
        return []

    if unshorten_ingka_pagelinks:
        if isinstance(item_codes, str):
            item_codes = [item_codes]
        unshortened_links = _get_unshortened_links_from_ingka_pagelinks(item_codes[0])
        if unshortened_links:
            item_codes.extend(unshortened_links)

    raw_res: list[str] = re.findall(
        r"\d{3}[, .-]{0,2}\d{3}[, .-]{0,2}\d{2}", str(item_codes)
    )
    regex = re.compile(r"[^0-9]")
    # http://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-in-python-whilst-preserving-order
    return list(dict.fromkeys(regex.sub("", i) for i in raw_res))


def format_item_code(item_code: str) -> str | None:
    matches = parse_item_codes(item_code)
    if not matches:
        return
    item_code = matches[0]
    return item_code[0:3] + "." + item_code[3:6] + "." + item_code[6:8]



import sys
from typing import Any

from ikea_api import parse_item_codes

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal


class ItemCode(str):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v: Any):
        if isinstance(v, int):
            v = str(v)
        if isinstance(v, str):
            parsed_item_codes = parse_item_codes(v)
            if len(parsed_item_codes) != 1:
                raise ValueError("invalid item code format")
            return parsed_item_codes[0]
        raise TypeError("string required")


ItemType = Literal["ART", "SPR"]


def get_is_combination_from_item_type(item_type: ItemType):
    assert item_type in ("ART", "SPR"), "Item type should be ART or SPR"
    return item_type == "SPR"



from __future__ import annotations

import datetime
from typing import List, Optional

from pydantic import BaseModel

from ikea_api._api import GraphQLResponse
from ikea_api.wrappers import types
from ikea_api.wrappers._parsers import translate_from_dict

__all__ = ["parse_status_banner_order", "parse_costs_order", "parse_history"]

STORE_NAMES = {"ru": {"IKEA": "Интернет-магазин", "Санкт-Петербург: Парнас": "Парнас"}}


class DateAndTime(BaseModel):
    date: datetime.date


class DeliveryDate(BaseModel):
    estimatedFrom: DateAndTime


class DeliveryMethod(BaseModel):
    deliveryDate: DeliveryDate


class StatusBannerOrder(BaseModel):
    dateAndTime: DateAndTime
    deliveryMethods: List[DeliveryMethod]


class StatusBannerData(BaseModel):
    order: StatusBannerOrder


class ResponseStatusBanner(BaseModel):
    data: StatusBannerData


class Cost(BaseModel):
    value: int


class CostsOrderCosts(BaseModel):
    delivery: Cost
    total: Cost


class CostsOrder(BaseModel):
    costs: CostsOrderCosts


class CostsData(BaseModel):
    order: CostsOrder


class ResponseCosts(BaseModel):
    data: CostsData


class HistoryDateAndTime(BaseModel):
    date: str
    time: str
    formattedLongDateTime: str


class HistoryTotalCost(BaseModel):
    value: Optional[int]


class HistoryItem(BaseModel):
    id: str
    status: str
    storeName: str
    dateAndTime: HistoryDateAndTime
    totalCost: HistoryTotalCost


class HistoryData(BaseModel):
    history: List[HistoryItem]


class ResponseHistory(BaseModel):
    data: HistoryData


def parse_status_banner_order(response: GraphQLResponse):
    order = ResponseStatusBanner(**response)  # type: ignore
    return types.StatusBannerOrder(
        purchase_date=order.data.order.dateAndTime.date,
        delivery_date=order.data.order.deliveryMethods[
            0
        ].deliveryDate.estimatedFrom.date,
    )


def parse_costs_order(response: GraphQLResponse):
    order = ResponseCosts(**response)  # type: ignore
    costs = order.data.order.costs
    return types.CostsOrder(
        delivery_cost=costs.delivery.value, total_cost=costs.total.value
    )


def get_history_datetime(item: HistoryItem):
    return f"{item.dateAndTime.date}T{item.dateAndTime.time}"


def parse_history(response: GraphQLResponse):
    history = ResponseHistory(**response)  # type: ignore
    return [
        types.PurchaseHistoryItem(
            id=i.id,
            status=i.status,
            price=i.totalCost.value or 0,
            datetime=get_history_datetime(i),
            datetime_formatted=i.dateAndTime.formattedLongDateTime,
            store=translate_from_dict(STORE_NAMES, i.storeName),
        )
        for i in history.data.history
    ]



from __future__ import annotations

import sys
from typing import Any

from ikea_api._api import CustomResponse, GraphQLAPI, GraphQLResponse
from ikea_api._constants import Constants
from ikea_api.exceptions import GraphQLError

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal


class Purchases(GraphQLAPI):
    def __init__(self, token: str):
        super().__init__(
            endpoint="https://purchase-history.ocp.ingka.ikea.com/graphql", token=token
        )
        self._session.headers.update(
            {
                "Accept": "*/*",
                "Accept-Language": f"{Constants.LANGUAGE_CODE}-{Constants.COUNTRY_CODE}",
                "Origin": "https://order.ikea.com",
                "Referer": f"https://order.ikea.com/{Constants.COUNTRY_CODE}/{Constants.LANGUAGE_CODE}/purchases/",
            }
        )

    def _build_payload(self, operation_name: str, query: str, **variables: Any):
        return {"operationName": operation_name, "variables": variables, "query": query}

    def history(self, *, take: int = 5, skip: int = 0):
        """Get purchase history.
        Parameters are for pagination. If you want to see all your purchases set 'take' to 10000.
        """
        payload = self._build_payload("History", Queries.history, take=take, skip=skip)
        return self._post(json=payload)

    def _error_handler(self, response: CustomResponse):
        if isinstance(response._json, list):  # from order_info
            dict_: GraphQLResponse
            for dict_ in response._json:  # type: ignore
                if "errors" in dict_:
                    raise GraphQLError(response)

    def order_info(
        self,
        order_number: str,
        *,
        email: str | None = None,
        queries: list[
            Literal["StatusBannerOrder", "CostsOrder", "ProductListOrder"]
        ] = ["StatusBannerOrder", "CostsOrder", "ProductListOrder"],
        skip_products: int = 0,
        skip_product_prices: bool = False,
        take_products: int = 10,
    ) -> list[GraphQLResponse]:
        """Get order information: status and costs.

        :params order_number: Purchase ID
        :params email: Email. If set, there's no need to get token.
        :params queries: Queries that will be included in request
        :params skip_products: Relevant to ProductListOrder
        :params skip_product_prices: Relevant to ProductListOrder
        :params take_products: Relevant to ProductListOrder
        """
        headers = {"Referer": f"{self._session.headers['Origin']}/{order_number}/"}
        payload: list[dict[str, Any]] = []

        if "StatusBannerOrder" in queries:
            payload.append(
                self._build_payload(
                    "StatusBannerOrder",
                    Queries.status_banner_order,
                    orderNumber=order_number,
                )
            )
        if "CostsOrder" in queries:
            payload.append(
                self._build_payload(
                    "CostsOrder",
                    Queries.costs_order,
                    orderNumber=order_number,
                )
            )
        if "ProductListOrder" in queries:
            payload.append(
                self._build_payload(
                    "ProductListOrder",
                    Queries.product_list_order,
                    orderNumber=order_number,
                    receiptNumber="",
                    skip=skip_products,
                    skipPrice=skip_product_prices,
                    take=take_products,
                )
            )

        if email:
            headers["Referer"] = self._session.headers["Referer"] + "lookup"
            for chunk in payload:
                chunk["variables"]["liteId"] = email

        return self._post(headers=headers, json=payload)  # type: ignore


class Fragments:
    date_and_time = """
    fragment DateAndTime on DateAndTime {
        time
        date
        formattedLocal
        formattedShortDate
        formattedLongDate
        formattedShortDateTime
        formattedLongDateTime
    }
    """

    delivery_date = """
    fragment DeliveryDate on DeliveryDate {
        actual {
        ...DateAndTime
        }
        estimatedFrom {
        ...DateAndTime
        }
        estimatedTo {
        ...DateAndTime
        }
        dateTimeRange
        timeZone
    }
    """

    service_info = """
    fragment ServiceInfo on Service {
        id
        status
        date {
        ...DeliveryDate
        }
    }
    """

    delivery_info = """
    fragment DeliveryInfo on DeliveryMethod {
        id
        serviceId
        status: status2
        type
        deliveryDate {
        ...DeliveryDate
        }
    }
    """

    money = """
    fragment Money on Money {
        code
        value
    }
    """

    tax_rate = """
    fragment TaxRate on TaxRate {
        percentage
        name
        amount {
        ...Money
        }
    }
    """

    costs = """
    fragment Costs on Costs {{
        total {{
        ...Money
        }}
        delivery {{
        ...Money
        }}
        service {{
        ...Money
        }}
        discount {{
        ...Money
        }}
        sub {{
        ...Money
        }}
        tax {{
        ...Money
        }}
        taxRates {{
        ...TaxRate
        }}
    }}
    {}
    {}
    """.format(
        money,
        tax_rate,
    )

    article = """
    fragment Article on Product {
        name
        description
        href
        quantity
        decimalQuantity
        priceUnitText
        id
        splitDelivery
        image {
            small
            medium
            large
        }
    }
    """

    product = """
    fragment Product on Product {
        ...Article
        unitPrice @skip(if: $skipPrice) {
            formatted
        }
        totalPrice @skip(if: $skipPrice) {
            formatted
        }
        assemblyRequired
    }
    %s
    """ % (
        article
    )


class Queries:
    history = """
    query History($skip: Int!, $take: Int!) {
        history(skip: $skip, take: $take) {
        id
        dateAndTime {
        ...DateAndTime
        }
        status
        storeName
            totalCost {
                code
                value
                formatted
            }
        }
    }

    %s
    """ % (
        Fragments.date_and_time
    )

    status_banner_order = """
    query StatusBannerOrder($orderNumber: String!, $liteId: String) {{
        order(orderNumber: $orderNumber, liteId: $liteId) {{
            id
            dateAndTime {{
            ...DateAndTime
            }}
            status
            services {{
            ...ServiceInfo
            }}
            deliveryMethods {{
                ...DeliveryInfo
            }}
        }}
    }}

    {}
    {}
    {}
    {}
    """.format(
        Fragments.date_and_time,
        Fragments.delivery_date,
        Fragments.service_info,
        Fragments.delivery_info,
    )

    costs_order = """
    query CostsOrder($orderNumber: String!, $liteId: String) {
        order(orderNumber: $orderNumber, liteId: $liteId) {
            id
            costs {
                ...Costs
            }
        }
    }

    %s

    """ % (
        Fragments.costs
    )

    product_list_order = """
    query ProductListOrder(
        $orderNumber: String!
        $liteId: String
        $skip: Int!
        $take: Int!
        $skipPrice: Boolean!
    ) {
        order(orderNumber: $orderNumber, liteId: $liteId) {
            id
            articles {
                ... on AnyDirectionProducts {
                    quantity
                    direction
                    any(skip: $skip, take: $take) {
                        ...Product
                    }
                }
                ... on ExchangeProducts {
                    numberOfInboundElements
                    numberOfOutboundElements
                    inboundQuantity
                    outboundQuantity
                    inbound(skip: $skip, take: $take) {
                    ...Product
                    }
                    outbound(skip: $skip, take: $take) {
                    ...Product
                    }
                }
            }
        }
    }

    %s
    """ % (
        Fragments.product
    )



from __future__ import annotations

from ikea_api._api import API
from ikea_api._constants import Constants
from ikea_api.exceptions import IKEAAPIError


class PipItem(API):
    def __init__(self):
        super().__init__(
            f"{Constants.BASE_URL}/{Constants.COUNTRY_CODE}/{Constants.LANGUAGE_CODE}/products"
        )
        self._session.headers["Accept"] = "*/*"

    def _request_item(self, item_code: str, is_combination: bool):
        prefix = "s" if is_combination else ""
        return self._get(f"{self.endpoint}/{item_code[5:]}/{prefix}{item_code}.json")

    def __call__(self, item_code: str):
        try:
            return self._request_item(item_code, True)
        except IKEAAPIError as exc:
            if not exc.response.status_code == 404:
                raise
            return self._request_item(item_code, False)



from __future__ import annotations

from typing import Any, List, Optional

from pydantic import BaseModel, HttpUrl

from ikea_api.wrappers import types
from ikea_api.wrappers._parsers.item_base import ItemCode

__all__ = ["main"]


class Catalog(BaseModel):
    name: str
    url: HttpUrl


class CatalogRef(BaseModel):
    elements: List[Catalog]


class CatalogRefs(BaseModel):
    products: Optional[CatalogRef]


class ResponsePipItem(BaseModel):
    id: ItemCode
    priceNumeral: int
    pipUrl: HttpUrl
    catalogRefs: CatalogRefs


def get_category_name_and_url(catalog_refs: CatalogRefs):
    if not catalog_refs.products:
        return None, None
    return catalog_refs.products.elements[0].name, catalog_refs.products.elements[0].url


def main(response: dict[str, Any]):
    if not response:
        return
    parsed_item = ResponsePipItem(**response)
    category_name, category_url = get_category_name_and_url(parsed_item.catalogRefs)
    return types.PipItem(
        item_code=parsed_item.id,
        price=parsed_item.priceNumeral,
        url=parsed_item.pipUrl,
        category_name=category_name,
        category_url=category_url,
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from requests import Response

if TYPE_CHECKING:
    from ikea_api._api import CustomResponse, GraphQLResponse


class IKEAAPIError(Exception):
    """Generic API related exception."""

    response: CustomResponse

    def __init__(self, response: CustomResponse, msg: Any = None):
        self.response = response
        if msg is None:
            msg = (response.status_code, response.text)
        super().__init__(msg)


class UnauthorizedError(IKEAAPIError):
    """Exception that is being called when cannot log in"""

    def __init__(self, response: CustomResponse):
        resp_json: dict[str, Any] = response._json
        msg = (
            resp_json.get("moreInformation")
            or resp_json.get("error")
            or resp_json.get("title")  # From Cart
            or resp_json
        )
        super().__init__(response, msg)


class GraphQLError(IKEAAPIError):
    """Generic GraphQL exception"""

    errors: list[dict[str, Any]] | dict[str, Any]

    def __init__(self, response: CustomResponse):
        resp: GraphQLResponse | list[GraphQLResponse] = response._json

        if isinstance(resp, dict):
            self.errors = resp["errors"]  # type: ignore
        else:
            # from purchases.order_info
            self.errors = [d["errors"] for d in resp if "errors" in d]  # type: ignore

        super().__init__(response, self.errors)


class ItemFetchError(IKEAAPIError):
    response: Response  # type: ignore

    def __init__(self, response: Response, msg: Any = None):
        super().__init__(response, msg=msg)  # type: ignore


class ParsingError(Exception):
    pass



from __future__ import annotations

import sys
from json.decoder import JSONDecodeError
from typing import Any

import requests

from ikea_api._constants import DEFAULT_HEADERS
from ikea_api.exceptions import GraphQLError, IKEAAPIError, UnauthorizedError

if sys.version_info < (3, 8):
    from typing_extensions import TypedDict
else:
    from typing import TypedDict


class CustomResponse(requests.Response):
    _json: Any


class API:
    """Generic API class"""

    def __init__(self, endpoint: str):
        self.endpoint = endpoint
        self._session = requests.Session()
        self._session.headers.update(DEFAULT_HEADERS)

    def _basic_error_handler(self, response: CustomResponse):
        if response.status_code == 401:  # Token did not passed
            raise UnauthorizedError(response)

    def _error_handler(self, response: CustomResponse):  # pragma: no cover
        pass

    def _handle_response(self, response: CustomResponse):
        try:
            resp_json = response.json()
        except JSONDecodeError:
            raise IKEAAPIError(response)
        response._json = resp_json
        self._basic_error_handler(response)
        self._error_handler(response)
        if not response.ok:
            raise IKEAAPIError(response)
        return response._json

    def _get(
        self,
        endpoint: str | None = None,
        headers: dict[str, str] | None = None,
        params: dict[str, Any] | None = None,
    ) -> Any:
        if not endpoint:
            endpoint = self.endpoint
        response: CustomResponse = self._session.get(  # type: ignore
            url=endpoint, headers=headers, params=params
        )
        return self._handle_response(response)

    def _post(
        self,
        endpoint: str | None = None,
        headers: dict[str, str] | None = None,
        json: Any = None,
    ) -> Any:
        if not endpoint:
            endpoint = self.endpoint
        response: CustomResponse = self._session.post(  # type: ignore
            url=endpoint, headers=headers, json=json
        )
        return self._handle_response(response)


class AuthorizedAPI(API):
    def __init__(self, endpoint: str, token: str | None):
        super().__init__(endpoint)
        self._token = token
        if token is not None:
            self._session.headers["Authorization"] = "Bearer " + token

    @property
    def token(self):
        if not self._token:
            raise RuntimeError("No token provided")
        return self._token


class GraphQLResponse(TypedDict):
    data: dict[str, Any]
    errors: list[dict[str, Any]] | None


class GraphQLAPI(AuthorizedAPI):
    def _basic_error_handler(self, response: CustomResponse):
        super()._basic_error_handler(response)
        if "errors" in response._json:
            raise GraphQLError(response)

    def _post(
        self,
        endpoint: str | None = None,
        headers: dict[str, str] | None = None,
        json: Any = None,
    ) -> GraphQLResponse:  # pragma: no cover
        return super()._post(endpoint=endpoint, headers=headers, json=json)



from __future__ import annotations

from typing import Any

from ikea_api._api import GraphQLAPI
from ikea_api._constants import Constants, Secrets


class Cart(GraphQLAPI):
    def __init__(self, token: str):
        super().__init__(endpoint="https://cart.oneweb.ingka.com/graphql", token=token)
        self._session.headers["X-Client-Id"] = Secrets.cart_x_client_id

    def _call_api(self, query: str, **variables: Any):
        return self._post(
            json={
                "query": query,
                "variables": {"languageCode": Constants.LANGUAGE_CODE, **variables},
            }
        )

    def show(self):
        return self._call_api(Queries.cart)

    def clear(self):
        return self._call_api(Mutations.clear_items)

    def _make_templated_items(self, items: dict[str, int]):
        return [
            {"itemNo": item_code, "quantity": qty} for item_code, qty in items.items()
        ]

    def add_items(self, items: dict[str, int]):
        """
        Add items to cart.
        Required items list format: {'item_no': quantity, ...}
        """
        return self._call_api(
            Mutations.add_items, items=self._make_templated_items(items)
        )

    def update_items(self, items: dict[str, int]):
        """
        Replace quantity for given item to the new one.
        Required items list format: {'item_no': quantity, ...}
        """
        return self._call_api(
            Mutations.update_items, items=self._make_templated_items(items)
        )

    def copy_items(self, *, source_user_id: str):
        """Copy cart from another account"""
        return self._call_api(Mutations.copy_items, sourceUserId=source_user_id)

    def remove_items(self, item_codes: list[str]):
        """
        Remove items by item codes.
        """
        return self._call_api(Mutations.remove_items, itemNos=item_codes)

    def set_coupon(self, code: str):
        return self._call_api(Mutations.set_coupon, code=code)

    def clear_coupon(self):
        return self._call_api(Mutations.clear_coupon)


class Fragments:
    item_props = """
    fragment ItemProps on Item {
        itemNo
        quantity
        type
        fees {
            weee
            eco
        }
        isFamilyItem
        childItems {
            itemNo
        }
        regularPrice {
            unit {
                inclTax
                exclTax
                tax
                validFrom
                validTo
                isLowerPrice
                isTimeRestrictedOffer
                previousPrice {
                    inclTax
                    exclTax
                    tax
                }
            }
            subTotalExclDiscount {
                inclTax
                exclTax
                tax
            }
            subTotalInclDiscount {
                inclTax
                exclTax
                tax
            }
            subTotalDiscount {
                amount
            }
            discounts {
                code
                amount
                description
                kind
            }
        }
        familyPrice {
            unit {
                inclTax
                exclTax
                tax
                validFrom
                validTo
            }
            subTotalExclDiscount {
                inclTax
                exclTax
                tax
            }
            subTotalInclDiscount {
                inclTax
                exclTax
                tax
            }
            subTotalDiscount {
                amount
            }
            discounts {
                code
                amount
                description
                kind
            }
        }
    }
    """

    product_props = """
    fragment ProductProps on Product {
        name
        globalName
        isNew
        category
        description
        isBreathTaking
        formattedItemNo
        displayUnit {
            type
            imperial {
                unit
                value
            }
            metric {
                unit
                value
            }
        }
        unitCode
        measurements {
            metric
            imperial
        }
        technicalDetails {
            labelUrl
        }
        images {
            url
            quality
            width
        }
    }
    """

    totals = """
    fragment Totals on Cart {
        regularTotalPrice {
            totalExclDiscount {
                inclTax
                exclTax
                tax
            }
            totalInclDiscount {
                inclTax
                exclTax
                tax
            }
            totalDiscount {
                amount
            }
            totalSavingsDetails {
                familyDiscounts
            }
        }
        familyTotalPrice {
            totalExclDiscount {
                inclTax
                exclTax
                tax
            }
            totalInclDiscount {
                inclTax
                exclTax
                tax
            }
            totalDiscount {
                amount
            }
            totalSavingsDetails {
                familyDiscounts
            }
            totalSavingsInclDiscount {
                amount
            }
        }
    }
    """

    cart_props = """
    fragment CartProps on Cart {{
        currency
        checksum
        context {{
            userId
            isAnonymous
            retailId
        }}
        coupon {{
            code
            validFrom
            validTo
            description
        }}
        items {{
        ...ItemProps
            product {{
            ...ProductProps
            }}
        }}
        ...Totals
    }}
    {}
    {}
    {}
    """.format(
        item_props,
        product_props,
        totals,
    )


class Mutations:
    add_items = """
    mutation AddItems(
        $items: [AddItemInput!]!
        $languageCode: String
    ) {
        addItems(items: $items, languageCode: $languageCode) {
        quantity
            context {
                userId
                isAnonymous
                retailId
            }
        }
    }
    """

    clear_coupon = """
    mutation ClearCoupon(
        $languageCode: String
    ) {
        clearCoupon(languageCode: $languageCode) {
        ...CartProps
        }
    }

    %s
    """ % (
        Fragments.cart_props
    )

    clear_items = """
    mutation ClearItems(
        $languageCode: String
    ) {
        clearItems(languageCode: $languageCode) {
        ...CartProps
        }
    }

    %s
    """ % (
        Fragments.cart_props
    )

    copy_items = """
    mutation CopyItems(
        $sourceUserId: ID!
        $languageCode: String
    ) {
        copyItems(sourceUserId: $sourceUserId, languageCode: $languageCode) {
        ...CartProps
        }
    }

    %s
    """ % (
        Fragments.cart_props
    )

    remove_items = """
    mutation RemoveItems(
        $itemNos: [ID!]!
        $languageCode: String
    ) {
        removeItems(itemNos: $itemNos, languageCode: $languageCode) {
        ...CartProps
        }
    }

    %s
    """ % (
        Fragments.cart_props
    )

    set_coupon = """
    mutation SetCoupon(
        $code: String!
        $languageCode: String
    ) {
        setCoupon(code: $code, languageCode: $languageCode) {
        ...CartProps
        }
    }

    %s
    """ % (
        Fragments.cart_props
    )

    update_items = """
    mutation UpdateItems(
        $items: [UpdateItemInput!]!
        $languageCode: String
    ) {
        updateItems(items: $items, languageCode: $languageCode) {
        ...CartProps
        }
    }

    %s
    """ % (
        Fragments.cart_props
    )


class Queries:
    cart = """
    query Cart(
        $languageCode: String
    ) {
        cart(languageCode: $languageCode) {
        ...CartProps
        }
    }
  %s
    """ % (
        Fragments.cart_props
    )



from __future__ import annotations

from typing import Any, List, Optional

try:
    from pydantic import BaseModel
except ImportError:
    raise RuntimeError(
        "To use wrappers you need Pydantic to be installed. "
        + "Run 'pip install \"ikea_api[wrappers]\"' to do so."
    )

from ikea_api import IKEA
from ikea_api._endpoints.item_ingka import IngkaItems
from ikea_api._endpoints.item_iows import IowsItems
from ikea_api._endpoints.item_pip import PipItem
from ikea_api._utils import parse_item_codes
from ikea_api.exceptions import GraphQLError, ItemFetchError
from ikea_api.wrappers import types
from ikea_api.wrappers._parsers import (
    item_ingka,
    item_iows,
    item_pip,
    order_capture,
    purchases,
)

__all__ = [
    "get_purchase_history",
    "get_purchase_info",
    "add_items_to_cart",
    "get_delivery_services",
    "get_items",
]


def get_purchase_history(api: IKEA) -> list[types.PurchaseHistoryItem]:
    response = api.purchases.history()
    return purchases.parse_history(response)


def get_purchase_info(
    api: IKEA, *, id: str, email: str | None = None
) -> types.PurchaseInfo:
    status_banner_resp, costs_resp = api.purchases.order_info(
        order_number=id, email=email, queries=["StatusBannerOrder", "CostsOrder"]
    )

    status_banner = purchases.parse_status_banner_order(status_banner_resp)
    costs = purchases.parse_costs_order(costs_resp)
    return types.PurchaseInfo(**status_banner.dict(), **costs.dict())


class _CartErrorExtensionsData(BaseModel):
    itemNos: List[str]


class _CartErrorExtensions(BaseModel):
    code: str
    data: Optional[_CartErrorExtensionsData]


class _CartError(BaseModel):
    extensions: _CartErrorExtensions


def add_items_to_cart(api: IKEA, *, items: dict[str, int]) -> types.CannotAddItems:
    api.cart.clear()
    pending_items = items.copy()
    cannot_add_items: list[str] = []

    while pending_items:
        try:
            api.cart.add_items(pending_items)
            break
        except GraphQLError as exc:
            errors: list[dict[str, Any]] = exc.errors  # type: ignore
            for error_dict in errors:
                error = _CartError(**error_dict)
                if error.extensions.code != "INVALID_ITEM_NUMBER":
                    continue
                if not error.extensions.data:
                    continue
                cannot_add_items += error.extensions.data.itemNos

            for item_code in cannot_add_items:
                pending_items.pop(item_code, None)

    return cannot_add_items


def get_delivery_services(
    api: IKEA, *, items: dict[str, int], zip_code: str
) -> types.GetDeliveryServicesResponse:
    cannot_add = add_items_to_cart(api, items=items)
    cannot_add_all_items = not set(items.keys()) ^ set(cannot_add)
    if cannot_add_all_items:
        return types.GetDeliveryServicesResponse(
            delivery_options=[], cannot_add=cannot_add
        )

    order_capture_api = api.order_capture(zip_code=zip_code)
    checkout = order_capture_api._get_checkout(
        order_capture_api._get_items_for_checkout()
    )
    service_area = order_capture_api._get_service_area(checkout)
    home_response = order_capture_api.get_home_delivery_services(
        (checkout, service_area)
    )
    collect_response = order_capture_api.get_collect_delivery_services(
        (checkout, service_area)
    )
    parsed_responses = order_capture.main(
        home_delivery_services_response=home_response,
        collect_delivery_services_response=collect_response,
    )
    return types.GetDeliveryServicesResponse(
        delivery_options=parsed_responses, cannot_add=cannot_add
    )


def _split_to_chunks(list_: list[Any], chunk_size: int):
    return (list_[i : i + chunk_size] for i in range(0, len(list_), chunk_size))


def _get_iows_items(item_codes: list[str]):
    responses: list[Any] = []
    fetcher = IowsItems()
    for item_codes_chunk in _split_to_chunks(item_codes, 90):
        try:
            responses += fetcher(item_codes_chunk)
        except ItemFetchError as exc:
            if "Wrong Item Code" not in exc.args[0]:
                raise
    return [item_iows.main(r) for r in responses]


def _get_ingka_items(item_codes: list[str]):
    responses: list[Any] = []
    fetcher = IngkaItems()
    for item_codes_chunk in _split_to_chunks(item_codes, 50):
        responses.append(fetcher(item_codes_chunk))
    res: list[types.IngkaItem] = []
    for resp in responses:
        res += item_ingka.main(resp)
    return res


def _get_pip_items(item_codes: list[str]):
    fetcher = PipItem()
    responses = (fetcher(i) for i in item_codes)
    return [item_pip.main(r) for r in responses]


def _get_pip_items_map(items: list[types.PipItem | None]):
    res: dict[str, types.PipItem] = {}
    for item in items:
        if item:
            res[item.item_code] = item
    return res


def _get_ingka_pip_items(item_codes: list[str]) -> list[types.ParsedItem]:
    ingka_items = _get_ingka_items(item_codes)
    pip_items = _get_pip_items(item_codes)
    pip_items_map = _get_pip_items_map(pip_items)

    res: list[types.ParsedItem] = []
    for ingka_item in ingka_items:
        pip_item = pip_items_map.get(ingka_item.item_code)
        if pip_item is None:
            continue
        res.append(
            types.ParsedItem(
                is_combination=ingka_item.is_combination,
                item_code=ingka_item.item_code,
                name=ingka_item.name,
                image_url=ingka_item.image_url,
                weight=ingka_item.weight,
                child_items=ingka_item.child_items,
                price=pip_item.price,
                url=pip_item.url,
                category_name=pip_item.category_name,
                category_url=pip_item.category_url,
            )
        )
    return res


def get_items(item_codes: list[str]) -> list[types.ParsedItem]:
    pending_item_codes = parse_item_codes(item_codes, unshorten_ingka_pagelinks=True)
    fetched_items_ingka_pip = _get_ingka_pip_items(pending_item_codes)

    fetched_item_codes = [i.item_code for i in fetched_items_ingka_pip]
    pending_item_codes = [i for i in pending_item_codes if i not in fetched_item_codes]
    if not pending_item_codes:
        return fetched_items_ingka_pip

    fetched_items_iows = _get_iows_items(pending_item_codes)
    return fetched_items_ingka_pip + fetched_items_iows



import json
import re
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel

from ikea_api import Constants
from ikea_api.wrappers import types
from ikea_api.wrappers._parsers.item_base import (
    ItemCode,
    ItemType,
    get_is_combination_from_item_type,
)

__all__ = ["main"]


class CatalogElement(BaseModel):
    CatalogElementName: Union[str, Dict[Any, Any]]
    CatalogElementId: Union[int, str, Dict[Any, Any]]


class CatalogElementList(BaseModel):
    CatalogElement: Union[CatalogElement, List[CatalogElement]]


class Catalog(BaseModel):
    CatalogElementList: CatalogElementList


class CatalogRefList(BaseModel):
    CatalogRef: Optional[Union[List[Catalog], Catalog]]


class Image(BaseModel):
    ImageUrl: Union[str, Dict[Any, Any]]
    ImageType: str
    ImageSize: str


class RetailItemImageList(BaseModel):
    RetailItemImage: List[Image]


class Measurement(BaseModel):
    PackageMeasureType: str
    PackageMeasureTextMetric: str


class RetailItemCommPackageMeasureList(BaseModel):
    RetailItemCommPackageMeasure: List[Measurement]


class GenericItem(BaseModel):
    ItemNo: ItemCode
    ProductName: str
    ProductTypeName: str


class ChildItem(GenericItem):
    Quantity: int
    # For `get_name`
    ItemMeasureReferenceTextMetric: None = None
    ValidDesignText: None = None
    RetailItemCommPackageMeasureList: Optional[RetailItemCommPackageMeasureList]


class RetailItemCommChildList(BaseModel):
    RetailItemCommChild: Union[List[ChildItem], ChildItem]


class Price(BaseModel):
    Price: int


class RetailItemCommPriceList(BaseModel):
    RetailItemCommPrice: Union[Price, List[Price]]


class ResponseIowsItem(GenericItem):
    ItemType: ItemType
    CatalogRefList: CatalogRefList
    ItemMeasureReferenceTextMetric: Optional[str]
    ValidDesignText: Optional[str]
    RetailItemCommPackageMeasureList: Optional[RetailItemCommPackageMeasureList]
    RetailItemImageList: RetailItemImageList
    RetailItemCommChildList: Optional[RetailItemCommChildList]
    RetailItemCommPriceList: RetailItemCommPriceList


def get_rid_of_dollars(d: Dict[str, Any]) -> Dict[str, Any]:
    return json.loads(
        re.sub(
            pattern=r'{"\$": ([^}]+)}',
            repl=lambda x: x.group(1) if x.group(1) else "",
            string=json.dumps(d),
        )
    )


def get_name(item: Union[ChildItem, ResponseIowsItem]):
    return ", ".join(
        part
        for part in (
            item.ProductName,
            item.ProductTypeName.capitalize(),
            item.ValidDesignText,
            item.ItemMeasureReferenceTextMetric,
        )
        if part
    )


def get_image_url(images: List[Image]) -> Optional[str]:
    # Filter images first in case no image with S5 size found
    images = [
        i
        for i in images
        if i.ImageType != "LINE DRAWING"
        and isinstance(i.ImageUrl, str)
        and i.ImageUrl.endswith((".png", ".jpg", ".PNG", ".JPG"))
    ]
    if not images:
        return

    for image in images:
        if image.ImageSize == "S5":
            return Constants.BASE_URL + image.ImageUrl  # type: ignore
    return Constants.BASE_URL + images[0].ImageUrl  # type: ignore


def parse_weight(v: str):
    matches = re.findall(r"[0-9.,]+", v)
    if matches:
        return float(matches[0])
    return 0.0


def get_weight(measurements: List[Measurement]):
    weight = 0.0
    if not measurements:
        return weight

    for m in measurements:
        if m.PackageMeasureType == "WEIGHT":
            weight += parse_weight(m.PackageMeasureTextMetric)
    return weight


def get_child_items(
    child_items: Union[List[ChildItem], ChildItem]
) -> List[types.ChildItem]:
    if not child_items:
        return []
    if isinstance(child_items, ChildItem):
        child_items = [child_items]

    return [
        types.ChildItem(
            name=get_name(child),
            item_code=child.ItemNo,
            weight=get_weight(
                child.RetailItemCommPackageMeasureList.RetailItemCommPackageMeasure
                if child.RetailItemCommPackageMeasureList
                else []
            ),
            qty=child.Quantity,
        )
        for child in child_items
    ]


def get_price(prices: Union[Price, List[Price]]):
    if not prices:
        return 0
    if isinstance(prices, list):
        return min(p.Price for p in prices)
    return prices.Price


def get_url(item_code: str, is_combination: bool):
    suffix = "s" if is_combination else ""
    return (
        f"{Constants.BASE_URL}/{Constants.COUNTRY_CODE}/"
        + f"{Constants.LANGUAGE_CODE}/p/-{suffix}{item_code}"
    )


def get_category_name_and_url(catalogs: Optional[Union[List[Catalog], Catalog]]):
    if not catalogs:
        return None, None
    if isinstance(catalogs, Catalog):
        catalogs = [catalogs]

    idx = 0 if len(catalogs) == 1 else 1
    category = catalogs[idx].CatalogElementList.CatalogElement
    if not category:
        return None, None
    if isinstance(category, list):
        category = category[0]
    if isinstance(category.CatalogElementName, dict) or isinstance(
        category.CatalogElementId, dict
    ):
        return None, None
    return (
        category.CatalogElementName,
        f"{Constants.BASE_URL}/{Constants.COUNTRY_CODE}/"
        + f"{Constants.LANGUAGE_CODE}/cat/-{category.CatalogElementId}",
    )


def main(response: Dict[str, Any]):
    response = get_rid_of_dollars(response)
    item = ResponseIowsItem(**response)

    is_combination = get_is_combination_from_item_type(item.ItemType)
    weight = (
        get_weight(item.RetailItemCommPackageMeasureList.RetailItemCommPackageMeasure)
        if item.RetailItemCommPackageMeasureList
        else 0.0
    )
    child_items = (
        item.RetailItemCommChildList.RetailItemCommChild
        if item.RetailItemCommChildList
        else []
    )
    category_name, category_url = get_category_name_and_url(
        item.CatalogRefList.CatalogRef
    )

    return types.ParsedItem(
        is_combination=is_combination,
        item_code=item.ItemNo,
        name=get_name(item),
        image_url=get_image_url(item.RetailItemImageList.RetailItemImage),
        weight=weight,
        child_items=get_child_items(child_items),
        price=get_price(item.RetailItemCommPriceList.RetailItemCommPrice),
        url=get_url(item.ItemNo, is_combination),
        category_name=category_name,
        category_url=category_url,  # type: ignore
    )



from __future__ import annotations

import sys
from typing import Any

from ikea_api._api import API
from ikea_api._constants import Constants

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

SearchType = Literal["PRODUCT", "CONTENT", "PLANNER", "REFINED_SEARCHES", "ANSWER"]


class Search(API):
    def __init__(self):
        super().__init__(
            f"https://sik.search.blue.cdtapps.com/{Constants.COUNTRY_CODE}/{Constants.LANGUAGE_CODE}/search-result-page"
        )

    def __call__(
        self,
        query: str,
        *,
        limit: int = 24,
        types: list[SearchType] = ["PRODUCT"],
    ) -> dict[str, dict[str, Any] | list[Any]]:
        return self._get(
            params={
                "autocorrect": "true",
                "subcategories-style": "tree-navigation",
                "types": ",".join(types),
                "q": query,
                "size": limit,
                "c": "sr",  # API client: sr - search results, sb - search bar
                "v": "20210322",  # API version
            },
        )



from __future__ import annotations

from typing import Any

from requests import Response

from ikea_api._api import API
from ikea_api._constants import Constants, Secrets
from ikea_api.exceptions import ItemFetchError


class IowsItems(API):
    """IOWS Item API. Works only for Russian market"""

    def __init__(self):
        super().__init__("https://iows.ikea.com/retail/iows/ru/ru/catalog/items/")
        self._session.headers.update(
            {
                "Accept": "application/vnd.ikea.iows+json;version=2.0",
                "Referer": f"{Constants.BASE_URL}/ru/ru/shoppinglist/",
                "consumer": Secrets.item_iows_consumer,
                "contract": Secrets.item_iows_contract,
            }
        )

    def _set_initial_items(self, item_codes: list[str]):
        self.items = {i: False for i in item_codes}  # item_code: is_combination

    def _build_payload(self, items: dict[str, bool]):
        return ";".join(
            f"{'SPR' if is_combintaion else 'ART'},{item_code}"
            for item_code, is_combintaion in items.items()
        )

    def _get_response(self, relapse: int = 0) -> Response:
        endpoint = f"{self.endpoint}{self._build_payload(self.items)}"
        response = self._session.get(endpoint)

        if (
            response.status_code == 404
        ):  # 404 only happens when something is really wrong or when it is wrong item code
            if len(self.items) == 1:
                if relapse == 0:
                    # Single item code with wrong item type
                    item_code = list(self.items.keys())[0]
                    self.items[item_code] = not self.items[item_code]
                    return self._get_response(relapse + 1)
                else:
                    raise ItemFetchError(response, "Wrong Item Code")

        if not response.text:
            return response

        return self._handle_response(response, relapse)

    def _handle_response(self, response: Response, relapse: int) -> Response:  # type: ignore
        r_json = response.json()

        if "ErrorList" not in r_json or relapse == 2:
            return response

        errors: list[dict[str, Any]] | dict[str, Any] = r_json["ErrorList"]["Error"]
        if isinstance(errors, dict):
            errors = [errors]

        for err in errors:
            if err.get("ErrorCode", {}).get("$") != 1101:
                # Not error with item type
                raise ItemFetchError(response, err)

            err_list: dict[str, Any] = {
                attr["Name"]["$"]: attr["Value"]["$"]
                for attr in err["ErrorAttributeList"]["ErrorAttribute"]
            }
            item_code = str(err_list["ITEM_NO"])
            if relapse == 0:  # Probably wrong item type: try to switch it
                self.items[item_code] = err_list["ITEM_TYPE"] == "ART"
            else:  # Nope, item is invalid
                del self.items[item_code]

        return self._get_response(relapse + 1)

    def __call__(self, item_codes: list[str]) -> list[Any]:
        if len(item_codes) > 90:
            raise RuntimeError("Can't get more than 90 items at once")

        self._set_initial_items(item_codes)
        resp = self._get_response()

        if not resp.text:
            return []

        data: dict[str, Any] = resp.json()

        if "RetailItemCommList" in data:
            return data["RetailItemCommList"]["RetailItemComm"]
        return [data["RetailItemComm"]]



from __future__ import annotations

import re
from typing import Any

from ikea_api._api import AuthorizedAPI
from ikea_api._constants import Constants, Secrets
from ikea_api._endpoints.cart import Cart


class OrderCapture(AuthorizedAPI):
    def __init__(self, token: str, *, zip_code: str, state_code: str | None = None):
        _validate_zip_code(zip_code)
        _validate_state_code(state_code)
        self._zip_code = zip_code
        self._state_code = state_code

        host = "ikea.ru" if Constants.COUNTRY_CODE == "ru" else "ingka.com"
        super().__init__(
            endpoint=f"https://ordercapture.{host}/ordercaptureapi/{Constants.COUNTRY_CODE}",
            token=token,
        )
        self._session.headers["X-Client-Id"] = Secrets.order_capture_x_client_id

    def _get_items_for_checkout(self) -> list[dict[str, str | int]]:
        cart = Cart(self.token).show()
        if not cart["data"].get("cart", {}).get("items"):
            raise RuntimeError("Cannot get items for Order Capture")
        return [
            {
                "quantity": item["quantity"],
                "itemNo": item["itemNo"],
                "uom": item["product"]["unitCode"],
            }
            for item in cart["data"]["cart"]["items"]
        ]

    def _get_checkout(self, items: list[dict[str, str | int]]) -> str:
        """Generate checkout for items"""
        resp: dict[str, str] = self._post(
            f"{self.endpoint}/checkouts",
            headers={"X-Client-Id": Secrets.order_capture_checkout_x_client_id},
            json={
                "channel": "WEBAPP",
                "checkoutType": "STANDARD",
                "shoppingType": "ONLINE",
                "deliveryArea": None,
                "items": items,
                "languageCode": Constants.LANGUAGE_CODE,
                "serviceArea": None,
                "preliminaryAddressInfo": None,
            },
        )
        if "resourceId" not in resp:
            raise RuntimeError("No resourceId for checkout")
        return resp["resourceId"]

    def _get_service_area(self, checkout: str) -> str:
        """Generate delivery area for checkout from Zip Code and State Code"""
        data = {"zipCode": self._zip_code}
        if self._state_code is not None:
            data["stateCode"] = self._state_code  # TODO: Test state code
        resp: dict[str, str] = self._post(
            f"{self.endpoint}/checkouts/{checkout}/service-area", json=data
        )
        if "id" not in resp:
            raise RuntimeError("No id for service area")
        return resp["id"]

    def get_home_delivery_services(
        self, checkout_and_service_area: tuple[str, str] | None = None
    ) -> dict[str, Any]:
        """Get available home delivery services"""
        if checkout_and_service_area:
            checkout, service_area = checkout_and_service_area
        else:
            checkout = self._get_checkout(self._get_items_for_checkout())
            service_area = self._get_service_area(checkout)

        return self._get(
            f"{self.endpoint}/checkouts/{checkout}/service-area/{service_area}/home-delivery-services"
        )

    def get_collect_delivery_services(
        self, checkout_and_service_area: tuple[str, str] | None = None
    ) -> dict[str, Any]:
        """Get available collect delivery services"""
        if checkout_and_service_area:
            checkout, service_area = checkout_and_service_area
        else:
            checkout = self._get_checkout(self._get_items_for_checkout())
            service_area = self._get_service_area(checkout)

        return self._get(
            f"{self.endpoint}/checkouts/{checkout}/service-area/{service_area}/collect-delivery-services"
        )


def _validate_zip_code(zip_code: str):
    if len(re.findall(r"[^0-9]", zip_code)) > 0:
        raise ValueError(f"Invalid zip code: {zip_code}")


def _validate_state_code(state_code: str | None):
    if state_code is not None:
        if len(state_code) != 2 or len(re.findall(r"[^A-z]+", state_code)) > 0:
            raise ValueError(f"Invalid state code: {state_code}")



from __future__ import annotations

from ikea_api._api import API, CustomResponse
from ikea_api._constants import Constants, Secrets
from ikea_api.exceptions import ItemFetchError


class IngkaStock(API):
    def __init__(self):
        super().__init__(
            f"https://api.ingka.ikea.com/cia/availabilities/{Constants.COUNTRY_CODE}/{Constants.LANGUAGE_CODE}"
        )
        self._session.headers.update(
            {
                "Accept": "*/*",
                "Referer": f"{Constants.BASE_URL}/{Constants.COUNTRY_CODE}/{Constants.LANGUAGE_CODE}/order/delivery/",
                "X-Client-Id": Secrets.stocks_ingka_x_client_id,
            }
        )

    def _error_handler(self, response: CustomResponse):
        if "errors" in response._json:
            try:
                item_code = [response._json["errors"][0]["details"]["itemNo"]]
            except (KeyError, TypeError, IndexError):
                item_code = None
            raise ItemFetchError(response, item_code)

    def __call__(self, item_codes: list[str]):
        return self._get(
            params={
                "itemNos": item_codes,
                "expand": "StoresList,Restocks,SalesLocations",
            }
        )



from __future__ import annotations

from typing import Any

from ikea_api._constants import Constants
from ikea_api._endpoints.auth import get_guest_token
from ikea_api._endpoints.cart import Cart
from ikea_api._endpoints.item_ingka import IngkaItems
from ikea_api._endpoints.item_iows import IowsItems
from ikea_api._endpoints.item_pip import PipItem
from ikea_api._endpoints.order_capture import OrderCapture
from ikea_api._endpoints.purchases import Purchases
from ikea_api._endpoints.search import Search, SearchType
from ikea_api._endpoints.stock_ingka import IngkaStock
from ikea_api._utils import format_item_code, parse_item_codes

__all__ = [
    "IKEA",
    "IngkaItems",
    "IowsItems",
    "PipItem",
    "format_item_code",
    "parse_item_codes",
    "IngkaStock",
]


class IKEA:
    token: str

    def __init__(
        self,
        token: str | None = None,
        *,
        country_code: str = "ru",
        language_code: str = "ru",
    ) -> None:
        self.token = token  # type: ignore
        Constants.COUNTRY_CODE = country_code
        Constants.LANGUAGE_CODE = language_code

    def login_as_guest(self) -> None:
        """Log in as guest. Token expires in 30 days."""
        self.token = get_guest_token()

    @property
    def cart(self) -> Cart:
        """Manage cart."""
        if not hasattr(self, "_cart"):
            self._cart = Cart(self.token)
        return self._cart

    def order_capture(
        self, *, zip_code: str, state_code: str | None = None
    ) -> OrderCapture:
        """Get available delivery services."""
        return OrderCapture(self.token, zip_code=zip_code, state_code=state_code)

    @property
    def purchases(self) -> Purchases:
        """Get information about your purchases."""
        if not hasattr(self, "_purchases"):
            self._purchases = Purchases(self.token)
        return self._purchases

    def search(
        self, query: str, *, limit: int = 24, types: list[SearchType] = ["PRODUCT"]
    ) -> dict[str, dict[str, Any] | list[Any]]:
        """Search the IKEA product catalog by product name"""
        return Search()(query=query, limit=limit, types=types)



import requests

from ikea_api._constants import DEFAULT_HEADERS, Constants, Secrets


def get_guest_token() -> str:
    resp = requests.post(
        url="https://api.ingka.ikea.com/guest/token",
        headers={
            **DEFAULT_HEADERS,
            "Accept": "*/*",
            "Accept-Language": "en-us",
            "X-Client-Id": Secrets.auth_guest_token_x_client_id,
            "X-Client-Secret": Secrets.auth_guest_token_x_client_secret,
        },
        json={"retailUnit": Constants.COUNTRY_CODE},
    )
    resp.reason = resp.text
    resp.raise_for_status()
    return resp.json()["access_token"]



from __future__ import annotations

import re
from typing import Any, List, Optional

from pydantic import BaseModel

from ikea_api._constants import Constants
from ikea_api.exceptions import ParsingError
from ikea_api.wrappers import types
from ikea_api.wrappers._parsers.item_base import (
    ItemCode,
    ItemType,
    get_is_combination_from_item_type,
)

__all__ = ["main"]


class ItemKey(BaseModel):
    itemType: ItemType
    itemNo: ItemCode


class PackageMeasurement(BaseModel):
    type: str
    valueMetric: float


class MediaVariant(BaseModel):
    quality: str
    href: str


class Media(BaseModel):
    typeName: str
    variants: List[MediaVariant]


class ProductType(BaseModel):
    name: str


class ValidDesign(BaseModel):
    text: str


class ReferenceMeasurements(BaseModel):
    metric: str


class Measurements(BaseModel):
    referenceMeasurements: Optional[List[ReferenceMeasurements]]


class LocalisedCommunication(BaseModel):
    languageCode: str
    packageMeasurements: Optional[List[PackageMeasurement]]
    media: Optional[List[Media]]
    productName: str
    productType: ProductType
    validDesign: Optional[ValidDesign]
    measurements: Optional[Measurements]


class ChildItem(BaseModel):
    quantity: int
    itemKey: ItemKey


class ResponseIngkaItem(BaseModel):
    itemKey: ItemKey
    localisedCommunications: List[LocalisedCommunication]
    childItems: Optional[List[ChildItem]]


class ResponseIngkaItems(BaseModel):
    data: List[ResponseIngkaItem]


def get_localised_communication(comms: list[LocalisedCommunication]):
    for comm in comms:
        if comm.languageCode == Constants.LANGUAGE_CODE:
            return comm
    raise ParsingError("Cannot find appropriate localized communication")


def _parse_russian_product_name(product_name: str):
    if not re.findall(r"[А-яЁё ]+", product_name):  # Russian text found
        # No russian text found: 'MARABOU'
        return product_name

    if re.findall(r"\d+", product_name):
        # Has numbers in itself: 'IKEA 365+ ИКЕА/365+', 'VINTER 2021 ВИНТЕР 2021'
        matches = re.findall(r"^[^А-яЁё]+?([А-яЁё].*)", product_name)
        if matches:
            product_name = matches[0]
    else:
        # Covers cases like: 'BESTÅ БЕСТО / EKET ЭКЕТ'
        product_name = re.sub(r"[^А-яЁё /+]+", "", product_name)
        product_name = re.sub(r"\s+", " ", product_name)
        product_name = re.sub(r"^ | $", "", product_name)

    return product_name


def get_name(comm: LocalisedCommunication):
    product_name = _parse_russian_product_name(comm.productName)
    product_type = comm.productType.name.capitalize()
    design = comm.validDesign.text if comm.validDesign else None

    if (
        comm.measurements
        and comm.measurements.referenceMeasurements
        and len(comm.measurements.referenceMeasurements) > 0
        and comm.measurements.referenceMeasurements[0].metric
    ):
        measurements = comm.measurements.referenceMeasurements[0].metric
    else:
        measurements = None

    return ", ".join(
        part for part in (product_name, product_type, design, measurements) if part
    )


def get_image_url(comm: LocalisedCommunication):
    if comm.media is None:
        return

    for media in comm.media:
        if media.typeName != "MAIN_PRODUCT_IMAGE":
            continue
        for v in media.variants:
            if v.quality == "S5":
                return v.href
    return comm.media[0].variants[0].href


def get_weight(comm: LocalisedCommunication):
    weight = 0.0
    if not comm.packageMeasurements:
        return weight

    for m in comm.packageMeasurements:
        if m.type == "WEIGHT":
            weight += m.valueMetric
    return weight


def get_child_items(child_items: list[ChildItem] | None) -> list[types.ChildItem]:
    if not child_items:
        return []

    return [
        types.ChildItem(
            item_code=child.itemKey.itemNo,
            name=None,
            weight=0.0,
            qty=child.quantity,
        )
        for child in child_items
    ]


def parse_item(item: ResponseIngkaItem):
    comm = get_localised_communication(item.localisedCommunications)
    return types.IngkaItem(
        is_combination=get_is_combination_from_item_type(item.itemKey.itemType),
        item_code=item.itemKey.itemNo,
        name=get_name(comm),
        image_url=get_image_url(comm),
        weight=get_weight(comm),
        child_items=get_child_items(item.childItems),
    )


def main(response: dict[str, Any]):
    parsed_resp = ResponseIngkaItems(**response)
    for item in parsed_resp.data:
        yield parse_item(item)



from __future__ import annotations

from typing import Any

from ikea_api._constants import Constants


def translate_from_dict(dictionary: dict[str, dict[str, Any]], v: str):
    lang_dict = dictionary.get(Constants.LANGUAGE_CODE)
    if lang_dict is None:
        return v
    return lang_dict.get(v, v)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ikea_api',
 'ikea_api._endpoints',
 'ikea_api.wrappers',
 'ikea_api.wrappers._parsers']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.26.0,<3.0.0']

extras_require = \
{':python_version < "3.9"': ['typing-extensions>=4.0.0,<5.0.0'],
 'test': ['pytest==7.0.1',
          'pytest-cov==3.0.0',
          'pytest-randomly==3.11.0',
          'responses==0.18.0'],
 'wrappers': ['pydantic>=1.8.2,<2.0.0']}

setup_kwargs = {
    'name': 'ikea-api',
    'version': '1.2.0',
    'description': "Client for several IKEA's APIs",
    'long_description': 'Client for several IKEA APIs.\n\n[![Test](https://github.com/vrslev/ikea-api-client/actions/workflows/test.yml/badge.svg)](https://github.com/vrslev/ikea-api-client/actions/workflows/test.yml)\n[![Coverage](https://codecov.io/gh/vrslev/ikea-api-client/branch/main/graph/badge.svg?token=Z1G75NBIC0)](https://codecov.io/gh/vrslev/ikea-api-client)\n[![Version](https://img.shields.io/github/v/release/vrslev/ikea-api-client?label=Version)](https://github.com/vrslev/ikea-api-client/releases/latest)\n[![Python](https://img.shields.io/pypi/pyversions/ikea_api?label=Python)](https://pypi.org/project/ikea_api)\n[![Downloads](https://img.shields.io/pypi/dm/ikea_api?label=Downloads)](https://pypi.org/project/ikea_api)\n[![License](https://img.shields.io/pypi/l/ikea_api?label=License)](https://github.com/vrslev/ikea-api-client/blob/main/LICENSE)\n\n# Features\n\nWith this library you can access the following IKEA\'s APIs:\n\n- Cart,\n- Delivery Services (actually, Order Capture),\n- Purchases (history and order info),\n- Items info (3 different services),\n- Search.\n\nAlso:\n\n- Fully typed and tested,\n- Has wrappers around most of APIs based on Pydantic.\n\n# Installation\n\n```bash\npip install ikea_api\n```\n\nIf you intend to use wrappers:\n\n```bash\npip install "ikea_api[wrappers]"\n```\n\n# Usage\n\n`IKEA` object unites all available the APIs that are in this package. This is done to share token, country and language.\n\n```python\nfrom ikea_api import IKEA\n\nikea = IKEA(token=None, country_code="ru", language_code="ru")\n```\n\n## Endpoints reference\n\n### 🔑 [Authorization](https://github.com/vrslev/ikea-api-client/blob/main/src/ikea_api/_endpoints/auth.py)\n\nFirst time you open IKEA.com, guest token is being generated and stored in cookies. Same thing must be done in here before using any endpoint.\n\nThis token expires in 30 days.\n\n```python\nikea.login_as_guest()\n```\n\nIt is stored in `token` property:\n\n```python\nikea.token  # Outputs JWT token\n```\n\nPreviously you could login as user (with login and password), but now there\'s very advanced telemetry that I wouldn\'t be able to solve in hundred years 🤪\n\n### 🛒 [Cart](https://github.com/vrslev/ikea-api-client/blob/main/src/ikea_api/_endpoints/cart.py)\n\nWith this endpoint you can do everything you can using IKEA\'s frontend:\n\n- Show the cart\n\n```python\nikea.cart.show()\n```\n\n- Clear it\n\n```python\nikea.cart.clear()\n```\n\n- Add, update and delete items\n\n```python\nikea.cart.add_items({"30457903": 1})  # { item_code: quantity }\n\nikea.cart.update_items({"30457903": 5})\n\nikea.cart.remove_items(["30457903"])\n```\n\n- Set and clear coupon\n\n```python\nikea.cart.set_coupon(...)\n\nikea.cart.clear_coupon()\n```\n\n- and even copy another user\'s cart.\n\n```python\nikea.cart.copy_items(source_user_id=...)\n```\n\nIf you use authorized token (copy-paste from cookies), than you edit your user\'s actual cart.\n\n> 💡\xa0There\'s wrapper that clears current cart and adds items with error handling: if requested item doesn\'t exist, the function just skips it and tries again.\n>\n> ```python\n> from ikea_api.wrappers import add_items_to_cart\n>\n> add_items_to_cart(  # Function returns items that can\'t be added. In this case: [\'11111111\']\n>     ikea,\n>     items={\n>         "30457903": 1,\n>         "11111111": 2,  # invalid item that will be skipped\n>     },\n> )\n> ```\n\n### 🚛 [Order Capture](https://github.com/vrslev/ikea-api-client/blob/main/src/ikea_api/_endpoints/order_capture.py)\n\nCheck pickup or delivery availability. If you need to know whether items are available _in stores_, use [Item availability endpoint](#%F0%9F%9F%A2-item-availabilityhttpsgithubcomvrslevikea-api-clientblobmainsrcikeaapiendpointsstockingkapy) or [ikea-availability-checker](https://github.com/Ephigenia/ikea-availability-checker).\n\n```python\norder = ikea.order_capture(\n    zip_code="02215",\n    state_code="MA",  # pass State Code only if your country has them\n)\nhome_delivery_services = order.get_home_delivery_services()\ncollect_delivery_services = order.get_collect_delivery_services()\n```\n\n> 💡\xa0You can use wrapper to add items to cart (clearing cart before and handling unknown item errors if they appear) and parse response in nice Pydantic models:\n>\n> ```python\n> from ikea_api.wrappers import get_delivery_services\n>\n> res = get_delivery_services(\n>    ikea,\n>    items={\n>        "30457903": 1,\n>        "11111111": 2,  # invalid item that will be skipped\n>    },\n>    zip_code="101000",\n> )\n> res.delivery_options  # List of parsed delivery services\n> res.cannot_add  # [\'11111111\']\n> ```\n\n### 📦 [Purchases](https://github.com/vrslev/ikea-api-client/blob/main/src/ikea_api/_endpoints/purchases.py)\n\n#### History\n\nThis method requires authentication, so if you don\'t have authorized token, it won\'t work.\n\n```python\nikea.purchases.history()\n\n# Get all purchases:\nikea.purchases.history(take=10000)\n\n# Pagination:\nikea.purchases.history(take=10, skip=1)\n```\n\n> 💡 Get parsed response with the wrapper:\n>\n> ```python\n> from ikea_api.wrappers import get_purchase_history\n>\n> get_purchase_history(ikea)  # Returns list of parsed purchases\n> ```\n\n#### Order info\n\n```python\nikea.purchases.order_info(order_number=..., email=...)\n\n# If you have authorized token, you can drop email:\nikea.purchases.order_info(order_number="111111111")\n\n# The method also has other params but they\'re mostly internal:\nikea.purchases.order_info(\n    order_number=...,\n    email=...,\n    queries=...,  # Queries that will be included in request, combine any of: ["StatusBannerOrder", "CostsOrder", "ProductListOrder"]. By default, all of them are included.\n    # Params below are relevant to ProductListOrder\n    skip_products=...,\n    skip_product_prices=...,\n    take_products=...,\n)\n```\n\n> 💡 Get parsed response with the wrapper:\n>\n> ```python\n> from ikea_api.wrappers import get_purchase_info\n>\n> get_purchase_info(  # Returns parsed purchase object. Items are not listed.\n>    ikea,\n>    id=...,\n>    email=...,\n> )\n> ```\n\n### 🪑 Item info\n\nGet item specification by item code (product number or whatever). There are 3 endpoints to do this because you can\'t get all the data about all the items using only one endpoint.\n\n```python\nfrom ikea_api import IowsItems, IngkaItems, PipItem\n\nitem_code = "30457903"\nitem_codes = [item_code]\n\n# <=90 items at a time\nIowsItems()([item_codes])\n\n# <=50 items at a time\nIngkaItems()([item_codes])\n\n# 1 item at a time\nPipItem()(item_code)\n```\n\n> 💡 You probably won\'t want to use raw APIs when there\'s convenient "smart" wrapper that combines them all and parses basic info:\n>\n> ```python\n> from ikea_api.wrappers import get_items\n>\n> get_items(["30457903"])\n> ```\n\n### 🟢 [Item availability](https://github.com/vrslev/ikea-api-client/blob/main/src/ikea_api/_endpoints/stock_ingka.py)\n\nGet availability by item code (product number or whatever).\n\n```python\nfrom ikea_api import IngkaStock\n\nitem_code = "30457903"\n\nIngkaStock()(item_code)\n```\n\n### 🔎 [Search](https://github.com/vrslev/ikea-api-client/blob/main/src/ikea_api/_endpoints/search.py)\n\nSearch for products in the product catalog by product name. Optionally also specify a maximum amount of returned search results (defaults to 24) and types of required search results.\n\n```python\nikea.search("Billy")\n\n# Retrieve 10 search results (default is 24)\nikea.search("Billy", limit=10)\n\n# Configure search results types\nikea.search(\n    "Billy",\n    types=...,  # Combine any of: ["PRODUCT", "CONTENT", "PLANNER", "REFINED_SEARCHES", "ANSWER"]\n)\n```\n\n### Utilities\n\n#### Parse item codes\n\nParse item codes from string or list.\n\n```python\nfrom ikea_api import parse_item_codes\n\nassert parse_item_codes("111.111.11") == ["11111111"]\nassert parse_item_codes("11111111, 222.222.22") == ["11111111", "22222222"]\nassert parse_item_codes("111") == []\n```\n\n#### Format item code\n\nParse item code and format it.\n\n```python\nfrom ikea_api import format_item_code\n\nassert format_item_code("11111111") == "111.111.11"\nassert format_item_code("111-111-11") == "111.111.11"\nassert format_item_code("111.111.11") == "111.111.11"\n```\n',
    'author': 'Lev Vereshchagin',
    'author_email': 'mail@vrslev.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/vrslev/ikea-api-client',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import sys

from ikea_api._api import API, CustomResponse
from ikea_api._constants import Constants, Secrets
from ikea_api.exceptions import ItemFetchError

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated


class IngkaItems(API):
    def __init__(self):
        super().__init__(
            f"https://api.ingka.ikea.com/salesitem/communications/{Constants.COUNTRY_CODE}/{Constants.LANGUAGE_CODE}"
        )
        self._session.headers.update(
            {
                "Accept": "*/*",
                "Referer": f"{Constants.BASE_URL}/{Constants.COUNTRY_CODE}/{Constants.LANGUAGE_CODE}/order/delivery/",
                "X-Client-Id": Secrets.item_ingka_x_client_id,
            }
        )

    def _error_handler(self, response: CustomResponse):
        if "error" in response._json:
            try:
                item_codes = response._json["error"]["details"][0]["value"]["keys"]
            except (KeyError, TypeError, IndexError):
                item_codes = None
            raise ItemFetchError(response, item_codes)

    def __call__(self, item_codes: Annotated[list[str], 50]):
        if len(item_codes) > 50:
            raise RuntimeError("Can't get more than 50 items at once")
        return self._get(params={"itemNos": item_codes})
